/*
Murasaki - multiple genome global alignment program
Copyright (C) 2006-2007 Keio University
(Kris Popendorf) <comp@bio.keio.ac.jp> (2006)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

/////////////////////
// murasaki config options manager
////////////////////

#include "options.h"
#include "murasaki.h"
#include "dinkymath.h"
#include <boost/filesystem/operations.hpp> //graceful handling of directories
#include <functional>
#include <vector>
#include <string>
#include <iostream>
#include <fstream>
#include <string.h>
#include <getopt.h>

#ifdef MURASAKI_MPI
#include "mpi.h"
#endif

using namespace std;
using namespace boost::filesystem;

Options::Options():
  quick_hash(3),hitfilter(1),histogram(1),
  user_seed(false),rand_seed(time(NULL)),repeatMask(false),
  skip1to1(false),skipFwd(false),skipRev(false),doubleHash(false),
  hashOnly(false),hashCache(false),bitscore(true),
  hashPref(EcoHash_t),alignment_name("test"),output_dir("output"),
  seedfilter(0),hashfilter(0),
  joinNeighbors(false),joinSpec(-1),joinDist(-1),
  hashSkip(1),
  auto_hashbits(true),
  dumpRegions(false),dumpEntropy(false)
{
  try {
    if(exists("~/.murasaki")){
	config_file="~/.murasaki";
	loadConfig(config_file); //load both!
    }
  }catch(exception &e){}
  try {
    if(exists("murasaki.cfg")){
      config_file="murasaki.cfg";
      loadConfig(config_file);//load both!
    }
  }catch(exception &e){}
}

void Options::commandline_opts(int argc,char **argv){
  int optc;
  while(1){
    //options struct:
    // name, has_arg, store_pointer, return_value
    static struct option long_options[] = {
      {"directory",1,0,'d'},
      {"pattern",1,0,'p'},
      {"name",1,0,'n'},
      {"help",0,0,'?'},
      {"version",0,0,'v'},
      {"quickhash",1,0,'q'},
      {"hashbits",1,0,'b'},
      {"hitfilter",1,0,'h'},
      {"histogram",2,0,'H'},
      {"hashtype",1,0,'t'},
      {"repeatmask",0,0,'r'},
      {"rseed",1,0,'s'},
      {"skip1to1",0,0,'1'},
      {"skipfwd",0,0,'F'},
      {"skiprev",0,0,'R'},
      {"doublehash",0,0,'D'},
      {"hashonly",0,0,'Q'},
      {"hashcache",2,0,'c'},
      {"seedfilter",1,0,'f'},
      {"hashfilter",1,0,'m'},
      {"nobitscore",0,0,'B'},
      {"join",2,0,'j'},
      {"hashskip",2,0,'S'},
      {"dumpregions",0,0,'G'},
      {"dumpentropy",0,0,'E'},
      {0,0,0,0}
    };
    int longindex=0;
    string prefreq;
    optc=getopt_long(argc,argv,"S:j:m:f:c::d:s:H::b:o:p:n:q:h:v?t:r1FR7DQB",long_options,&longindex);
    if(optc==-1)break;
    switch(optc){
    case 'j':
      if(optarg && (strcmp(optarg,"0"))){
	string optstr(optarg);
	if(optstr==string("0") || optstr==string("false")){
	  joinNeighbors=false;
	  break;
	}
      }
      joinNeighbors=true;
      joinSpec=-1;
      if(optarg && !sscanf(optarg,"%f",&joinSpec)){
	cerr << "Could not parse join specification.\n";
	cerr << program_help();exit(-1);
      }
      break;
    case 'S':
      if(optarg){
	if(!sscanf(optarg,"%d",&hashSkip)){
	  cerr << "Could not parse hashskip specification.\n";
	  cerr << program_help();exit(-1);
	}
      }else{
	hashSkip++;
      }
      if(hashSkip<=0){
	cerr << "Skip has to be at least 1.\n";
	exit(-1);
      }
      break;
    case 'c': 
      hashCache=true;
      if(optarg) cache_dir=string(optarg)+"/";
      else cache_dir="cache/";
      break;
    case 'v': cout << program_version();exit(-1);break;
    case 'd': output_dir=string(optarg);break;
    case 'p': pattern_src=string(optarg);break;
    case 'n': alignment_name=string(optarg);break;
    case 'q':
      if(!sscanf(optarg,"%d",&quick_hash)){
	cerr << "Could not parse quickhash specification.\n";
	cerr << program_help();exit(-1);
      }
      if(quick_hash>3 || quick_hash<0){
	cerr << "Invalid quickhash specification ("<<quick_hash<<")."<<endl;
	quick_hash=3;
	cerr << "Setting quickhash to "<<quick_hash<<" instead."<<endl;
      }
      break;
    case 'h':
      if(!sscanf(optarg,"%d",&hitfilter)){
	cerr << "Could not parse hitfilter specification.\n";
	cerr << program_help();exit(-1);
      }
      break;
    case 'b':
      if(!sscanf(optarg,"%d",&max_hash_bits)){
	cerr << "Could not parse hash size specification.\n";
	cerr << program_help();exit(-1);
      }
      if(max_hash_bits>WORDSIZE || max_hash_bits<1){
	cout << "*** Warning *** "<<max_hash_bits << " hash bits?\n";
	cout << "A hash that big would get you in trouble...\nI'm going to make that "<<(WORDSIZE-3)<<" for you... (and even that may be too big for your system)\n" << endl;
	max_hash_bits=(WORDSIZE-3);
      }
      auto_hashbits=false;
      break;
    case 'H':
      histogram=1;
      if(optarg && sscanf(optarg,"%d",&histogram) && (histogram<0 || histogram>4)){
	cerr << histogram << " is not a valid choice for histogram.\n";
	histogram=2;
	cerr << "Using "<<histogram<<" instead.\n";
      }
      break;
    case 't':
      prefreq=optarg;
      if(prefreq=="ArrayHash")
	hashPref=ArrayHash_t;
      else if(prefreq=="MSetHash" || prefreq=="MultisetHash")
	hashPref=MSetHash_t;
#ifdef MURASAKI_SQLITE3
      else if(prefreq=="SQLiteHash")
	hashPref=SQLiteHash_t;
#endif
#ifdef MURASAKI_MYSQL
      else if(prefreq=="MySQLHash")
	hashPref=MySQLHash_t;
#endif
      else if(prefreq=="EcoHash")
	hashPref=EcoHash_t;
      else cerr << "Unknown hash specification: "<<prefreq<<". Sticking with "<<hashPref_s()<<endl;
      break;
    case 'r':
      repeatMask=true;
      break;
    case 's':
      if(!sscanf(optarg,"%d",&rand_seed)){
	cerr << "Could not parse random seed specification (should be a number).\n";
	cerr << program_help();exit(-1);
      }
      user_seed=true;
      break;
    case 'f':
      if(!sscanf(optarg,"%d",&seedfilter)){
	cerr << "Could not parse seedfilter specification (should be a number).\n";
	cerr << program_help();exit(-1);
      }
      if(seedfilter<0)
	seedfilter=0;
      break;
    case 'm':
      if(!sscanf(optarg,"%d",&hashfilter)){
	cerr << "Could not parse seedfilter specification (should be a number).\n";
	cerr << program_help();exit(-1);
      }
      if(hashfilter<0)
	hashfilter=0;
      break;
    case 'B': bitscore=false;break;
    case 'F': skipFwd=true;break;
    case 'R': skipRev=true;break;
    case '1':case '7': skip1to1=true;break;
    case 'D': doubleHash=true;break;
    case 'Q': hashOnly=true;break;
    case 'G': dumpRegions=true;break;
    case 'E': dumpEntropy=true;break;
    default: cout << program_help();exit(1);break;
    }
  }
  if((seedfilter || hashfilter) && !doubleHash){
    cout << "Seedfilter/hashfilter necessitates doublehash: Enabling.\n";
    doubleHash=true;
  }
}

void Options::solidify(){ //finalize annoying names and crap
  if(joinNeighbors)
    if(joinSpec<0)
      joinDist=(int)(-1.0*joinSpec*(float)pattern_src.length());
    else
      joinDist=(int)joinSpec;

  if(hashCache){
    path cachedir(cache_dir);
    if(exists(cachedir)){
      if(!is_directory(cachedir)){
	cerr << "Cache directory is not a directory...\n";exit(-5);
      }
    }else{
      bool res=false;
      try{res=create_directory(cachedir);}
      catch(exception &e){cerr << "Exceptionated!!!\n"<<e.what()<<endl;exit(-5);}
      if(res)
	cout << "Created cache directory: "
	     <<system_complete(cachedir).string()<<endl;
    }
  }
  
  prefix=output_dir+"/"+alignment_name;
  path dir(output_dir);
  if(dir.empty())
    dir=current_path();
  else {
    if(exists(dir)){
      if(!is_directory(dir)){
	cerr << "Output directory is not a directory...\n";exit(-5);
      }
    }else{
      bool res=false;
      try{res=create_directory(dir);}
      catch(exception &e){cerr << "Exceptionated!!!\n"<<e.what()<<endl;exit(-5);}
      if(res)
	cout << "Created output directory: "<<system_complete(dir).string()<<endl;
    }
  }


#ifdef MURASAKI_MPI
  if( mpi_id!=0 ){
    char mpi_id_str[81];
    snprintf(mpi_id_str, 80, "%d", mpi_id);
    status_record=prefix+"."+mpi_id_str+".status";
  }
  else
#endif
    status_record=prefix+".status";

  hashStatus_record=prefix+".hashstatus";
  seq_record=prefix+".seqs";
  anchor_record=prefix+".anchors";
  pat_record=prefix+".pat";
  region_record=prefix+".regions";
  hashHisto_record=(prefix+".histogram");
  hashDetailed_record=(prefix+".histogram.details");
  options_record=prefix+".options";
  entropy_record=prefix+".entropy";
}

ostream& operator<<(ostream &opf,const Options &a){
  opf << "Writing to: " << a.prefix<<endl;
  opf << "Pattern is: " << pattern_src<<endl;
  opf << "Max usable hash bits: "<<max_hash_bits<<endl;
  opf << "Hit filter is: "<<a.hitfilter<<endl;
  opf << "Hash type is: "<<a.hashPref_s()<<endl;
  opf << "Double hash: "<<(a.doubleHash ? "on":"off")<<endl;
  opf << "Hashing every: "<<(a.hashSkip)<<" base"<<endl;
  opf << "Histogram detail is: "<<a.histogram<<endl;
  opf << "Hashing function is: "<<a.quick_hash<<endl;
  opf << "Repeats are: "<<(a.repeatMask ? "masked":"unmasked")<<endl;
  if(a.seedfilter)
    opf << "Seedfilter is at: "<<a.seedfilter<<endl;
  if(a.hashfilter)
    opf << "Hashfilter is at: "<<a.hashfilter<<endl;
  opf << "Random seed: "<<(a.rand_seed)<<(a.user_seed ? " (custom)":" (automatic)")<<endl;
  if(a.skipFwd || a.skipRev || a.skip1to1)
    opf << "Skip: "<<(a.skipFwd ? "forward ":"")<<(a.skipRev ? "reverse ":"")
	<< (a.skip1to1 ? "1-to-1 ":"") << endl;
  opf << "Bitscore output: "<<(a.bitscore ? "on":"off")<<endl;
  opf << "Join neighbors is: "<<(a.joinNeighbors ? (string(" at ")+dstring(a.joinDist)):"off")<<endl;
  opf << program_version();
  return opf;
}

string Options::hashPref_s() const{
  switch(hashPref){
  case ArrayHash_t:return string("Array");
  case MSetHash_t:return string("Multiset");
  case MySQLHash_t:return string("MySQL");
  case SQLiteHash_t: return string("SQLite");
  case EcoHash_t:return string("EcoHash");
  default: return string("Buh????");
  }
  return string("Unknown?????"); //i certainly hope not
}

void Options::loadConfig(string file){
  ifstream cfg(file.c_str());
  //construct a phony argv
  vector<char*> argv;
  argv.push_back("murasaki");//uhhh just dont read that, please
  string tmp;
  char buf[2048];
  while(cfg.good()){
    cfg.getline(buf,sizeof(buf));
    tmp=string(buf);
    while(cfg.fail()){ //for absurdly long inputs
      cfg.getline(buf,sizeof(buf));
      tmp+=buf;
    }
    argv.push_back(new char[tmp.length()]);
    strcpy(argv.back(),tmp.c_str());
  }
  commandline_opts(argv.size(),&argv.front()); //hoho stl rocks
}

string program_help(){
  return string("\
Usage: murasaki [options] seq1 seq2 [seq3 ... ]\n\
Options:\n\
  --directory|-d = output directory (default: output)\n\
  --name|-n      = alignment name (default: test)\n\
  --pattern|-p   = seed pattern (default: 111010010100110111)\n\
                    using the format [<w>:<l>] automatically generates a\n\
                    random pattern of weight <w> and length <l>\n\
  --quickhash|-q = specify a hashing function:\n\
                    1 - don't pack bits to make hash (use first word only)\n\
                    2 - naively use the first hashbits worth of pattern\n\
                    3 - adaptivevely find a good hash (default)\n\
  --hashbits|-b  = use n bit hashes (for n's of 1 to WORDSIZE. default 26)\n\
  --hashtype|-t  = select hash data structure to use:\n\
                    EcoHash    - Very tight packing of hashbits (experimental)\n\
                    ArrayHash  - memory frugal malloc/realloc (default)\n\
                    MSetHash   - memory exorbanant, but guaranteeably fast\n\
                    SQLiteHash - use SQLite3\n\
                    MySQLHash  - use MySQL\n\
  --hitfilter|-h = minimum number of hits to be outputted as an anchor\n\
                   (default 1)\n\
  --histogram|-H = histogram computation level: (-H alone implies -H1)\n\
                    0 - no histogram\n\
                    1 - basic bucketsize/bucketcount histogram data (def)\n\
		    2 - bucket-based scores to anchors.detils\n\
                    3 - perbucket count data\n\
		    4 - perbucket + perpattern count data\n\
  --repeatmask|-r= skip repeat masked data (ie: lowercase atgc)\n\
  --seedfilter|-f= skip seeds that occur more than N times\n\
                   ie. --seedfilter=1 would cause Mauve-like behaviour\n\
                   ie. only seeds which occur once per sequence are used\n\
  --hashfilter|-m= like --seedfilter but works on hash keys instead of\n\
                   seeds. May cause some collateral damage to otherwise\n\
                   unique seeds, but it's faster. Also non-sequence-specific\n\
                   so more like at best 1/N the tolerance of seedfilter.\n\
  --rseed|-s     = random number seed for non-deterministic algorithms\n\
                   (ie: the adative hash-finding). If you're doing any\n\
                   performance comparisons, it's probably imperative that you\n\
                   use the same seed for each run of the same settings.\n\
                   Default is obtained from time() (ie: seconds since 1970).\n\
  --skipfwd|-F   = Skip forward facing matches\n\
  --skiprev|-R   = Skip reverse facing matches\n\
  --skip1to1|-1  = Skip matches along the 1:1 line (good for comparing to self)\n\
  --doublehash|-D= Hashes the final sequence twice (once to fill the hash,\n\
                   and once to evaluate matches on it). This is good if your\n\
                   hash happens to be stored externally in a database.\n\
  --hashonly|-Q  = Hash Only. no anchors. just statistics.\n\
  --hashskip|-S  = Hashes every n bases. (Default is 1. ie all)\n\
                   Not supplying any argument increments the skip amount by 1.\n\
  --hashCache|-c = Caches hash tables in the directory. (default: cache/)\n\
  --join|-j      = Join anchors within n bases of eachother (default: 0)\n\
                   Specifying a negative n implies -n*patternLength\n\
  --nobitscore|-B= skip computing a bitscore for all anchors\n\
  --dumpentropy|-E = dump entropy data for each anchor\n\
  --dumpregions|-G = dump list of comparable regions for each sequence\n\
                     (ie: the areas between N's)\n\
\n\
...and of course\n\
  --version|-v   = prints version information and quits\n\
  --help|-?      = prints this help message and quits\n\
\n")+platformInfo()+program_version();
}
