/*
Murasaki - multiple genome global alignment program
Copyright (C) 2006-2007 Keio University
(Kris Popendorf) <comp@bio.keio.ac.jp> (2006)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

//////////////
// murasaki project
// sequence.cc
// provides "sequence" class for doing operations on dna sequences
//////////////

#include "murasaki.h"
#include "sequence.h"
#include "dinkymath.h"
#include "cmultiset.h"
#include "ecohash.h"

#include <pstream.h>
#include <fstream>
#include <limits.h>
#include <assert.h>
#include <set>
#include <iterator>
#include <functional>
#include <algorithm>
#include <stdlib.h> //need random()
#include <math.h> //need log...
#include <ctype.h> //for tolower <-- im a lazy lazy man

//ahh boosty my old friend...
#include <boost/spirit/core.hpp>
#include <boost/spirit/attribute.hpp>
#include <boost/spirit/symbols/symbols.hpp>
#include <boost/spirit/utility/chset.hpp>
#include <boost/spirit/utility/escape_char.hpp>
#include <boost/spirit/utility/confix.hpp>
#include <boost/spirit/iterator.hpp>
#include <boost/spirit/error_handling/exceptions.hpp>

#ifdef MURASAKI_SQLITE3
#include <sqlite3.h>
#endif

#ifdef MURASAKI_MYSQL
#include <mysql.h>
#endif

//these guys are used in creating the adaptive hash stuff:
#include <ext/functional>
using namespace __gnu_cxx;

using namespace std;


int next_sequence_=0;
int _wordsize=WORDSIZE;
int hash_bits=26; 
int max_hash_bits=26; //keeps the approx 256mb hash as default (or 515mb on 64bit)
word hash_mask=lowN(hash_bits);
int hash_max=intpow(2,max_hash_bits);
//int hash_max=67108879; //first prime over .5gb/8bytes
//524288057; //arbitrary default. first prime number past .5gb
int hash_size;

word alternating_mask;
word activeHash; //for building hash member lists

const int FIRST2=highN(2);

//static inits
map<int,AnchorSet*> AnchorSet::setMap;

int AnchorSet::_nextId=0;
int HashMethod::nextId=0;
map<int,int> HashMethod::popmap;

//seqtable bits
Sequence** seq_table;
SeqIdx seq_table_size;
map<Sequence*,SeqIdx> seq2idx;
bool _seq_no_data=false;
double globalBaseFreq[4]={0,0,0,0};
word globalBaseCount[4]={0,0,0,0};
word globalCounted=0;

void set_seqtable(vector<Sequence*> seqs){
  seq_table_size = seqs.size();

  seq_table = (Sequence**)malloc(sizeof(Sequence*)*seq_table_size);
  if(seq_table == NULL) { cerr << "set_seqtable(): malloc() failed." << endl; exit(-1); }
  for(SeqIdx i=0; i<seq_table_size; i++){
    seq_table[i] = seqs[i];
    seq2idx[seqs[i]] = i;
  }  
}

void initHashParams(int bits,BitSequence *pat){
  hash_bits=bits;
  hash_mask=lowN(bits);
  hash_size=intpow(2,bits);
  cout << "Using "<<bits<<" hashbits"<<endl;
  srandom(opt.rand_seed);
#ifndef NDEBUG
  srandom(0);
#endif
  if(opt.quick_hash==3)
    pat->compileHashFunc();
  if(opt.hashPref==EcoHash_t)
    EcoHash::init(seq_count,longestSeq,totalHashLength,bits);
}

void initConstants(){
  srandom(time(0));
  alternating_mask=1;
  //make a mask like 010101010101010101
  for(int bit=2;bit<WORDSIZE;bit<<=1)
    alternating_mask|= (alternating_mask<<bit);
}

int highestBit(word t){
  for(int i=WORDSIZE-4;i>=0;i-=4){
    switch((t>>i) & (word)15){
    case 1:return i;
    case 2:case 3:return i+1;
    case 4:case 5:case 6:case 7:return i+2;
    case 8:case 10:case 11:case 12:case 13:case 14:case 15:
      return i+3;
      //this could be unrolled arbitrarily for performance
    }
  }
  return -1;//wtf?
}

int lowestBit(word t){
  for(int i=0;i<WORDSIZE;i+=4){
    switch((t>>i) & (word)15){
    case 1:case 3:case 5:case 7: case 9: case 11: case 13: case 15:
      return i;
    case 2:case 6:case 10:case 14:return i+1;
    case 4:case 12:return i+2;
    case 8:return i+3;
      //this could be unrolled arbitrarily for performance
    }
  }
  return -1;//wtf?
}

int matchCount(word t){ //count whole-base matches
  int count=0;
  for(int i=0;i<WORDSIZE;i+=4){
    switch((t>>i) & (word)15){ //4 bits at a time
    case 3:case 12:count+=1;break;
    case 15:count+=2;break;
    default:break;
    }
  }
  return count;
}

int popCount(word t){
  int count=0;
  for(int i=0;i<WORDSIZE;i+=4){
    switch((t>>i) & (word)15){
    case 1:case 2:case 4:case 8:count+=1;break;
    case 3:case 5:case 6:case 9:case 10:case 12:count+=2;break;
    case 7:case 11:case 13:case 14:count+=3;break;
    case 15:count+=4;
      //this could be unrolled arbitrarily for performance
    }
  }
  return count;
}

int Sequence::length(){
  return fwd->length();
}

word Sequence::mclass(){
  return 1<<seqID;
}

Sequence::Sequence()
  : 
  fwd(),rev(),seqID(next_sequence_++)
{
}

Sequence::Sequence(string _filename):
  filename(_filename),name(_filename),
  seqID(next_sequence_++)  
{
  string cmd("./geneparse.pl");
  vector<string> argv(1);
  if(opt.repeatMask)
    argv.push_back(string("-r"));
  argv.push_back(string("-c"));
  argv.push_back(filename);
  cout << "Parsing genebank data from "<<filename<<endl;
  if(_seq_no_data){
    cout << "Initializing sequence "<<filename<<" without data.\n";
  }else{
    redi::ipstream inf(cmd,argv);
    string str;
    while(inf.good()){
      inf >> str; //well. we coudl be smarter, but this make revcomp simple
    }
    
    fwd=new BitSequence(str);
    cout << "Done ("<<str.length()<<"bp). Generating reverse complement"<<endl;
    reverse(str.begin(),str.end());
    rev=new BitSequence(str);
    rev->invert();
  }
  cout << "Done."<<endl;
}

Sequence::Sequence(const Sequence &a):
  filename(a.filename),
  name(a.name),
  fwd(a.fwd),
  rev(a.rev),  seqID(a.seqID)
{
}

//inline int Sequence::getId(){return seqID;}

//inline const vector<word>& BitSequence::words(){return _words;}
int BitSequence::wordsize(){return WORDSIZE;};

inline int BitSequence::readCode(int pos) const {
  word w=readWord(pos/WORDSIZE);
  return first2(w<<(MODWORDSIZE(pos)));
}

word BitSequence::readWord(int pos) const{
  if(pos>=word_count || pos<0)
    return 0;
  return _words[pos];
}

string bitToString(word w){
  switch(w & (word)3){
  case 0:return "C"; break;
  case 1:return "A"; break;
  case 2:return "T"; break;
  case 3:return "G"; break;
  }
  return " ";
}

string wordToString(word w,int bits){
  string ret;
  int start=bits < 0 ? bits+WORDSIZE:0;
  for(int i=start;i<WORDSIZE;i+=2)
    switch(first2(w<<i)){
    case 0:ret+="C"; break;
    case 1:ret+="A"; break;
    case 2:ret+="T"; break;
    case 3:ret+="G"; break;
    }
  return ret;
}

string wordToPattern(word w,int bits){
  string ret;
  for(int i=0;i<bits;i+=2)
    switch(first2(w<<i)){
    case 0:ret+="0"; break;
    case 1:ret+="-"; break;
    case 2:ret+="x"; break;
    case 3:ret+="1"; break;
    }
  return ret;
}

string BitSequence::asString() const {
  string ret;
  for(int i=0;i<word_count;i++){
    if(i%4==0){
      char buf[50];
      sprintf(buf,"%4d ",i*WORDSIZE/2+1);
      ret+=buf;
    }
    ret+=wordToString(readWord(i),i<word_count-1 ? WORDSIZE:
		      ((bit_count-1)%WORDSIZE)+1);
    if(i<word_count-1 && (i+1)%4==0)
	ret+="\n";
      else
	ret+=" ";
  }
  return ret;
}

string BitSequence::asPattern() const {
  string ret;
  for(int i=0;i<word_count;i++){
    if(i%(80/WORDSIZE*2)==0){ //used to be i%4 but that looks weird on hpc
      char buf[50];
      sprintf(buf,"%4d ",i*WORDSIZE/2+1);
      ret+=buf;
    }
    ret+=wordToPattern(readWord(i),i<word_count-1 ? WORDSIZE:
		      ((bit_count-1)%WORDSIZE)+1);
    if(i<word_count-1 && (i+1)%(80/WORDSIZE*2)==0)
	ret+="\n";
      else
	ret+=" ";
  }
  return ret;
}

int BitSequence::compHashLength(){
  int count=0;
  for(int i=0;i<word_count;i++){
    count+=popCount(readWord(i))/2;
  }
  _hashLength=count;
  return count;
}

BitSequence* BitSequence::substr(int start,int stop){
  assert(start<=stop);
  int pos=start*2,last=stop*2;//to bitspace;
  int myOffset=pos%WORDSIZE;
  word myMask=lowN(myOffset);
  int myFrame=pos/WORDSIZE;
  word myWord=readWord(myFrame),myNext=readWord(myFrame+1);
  BitSequence* ret=new BitSequence;
  while(pos+WORDSIZE<=last){
    if(myOffset)
      myWord=myWord<<myOffset | (myNext & myMask);
    ret->addWord(myWord,WORDSIZE);
    pos+=WORDSIZE;
    myFrame++;
    myWord=myNext;myNext=readWord(myFrame+1);
  }
  int remaining=last-pos;
  if(remaining>0){
    ret->addWord(myWord,remaining);
  }
  return ret;
}

//operator
int BitSequence::cmpRight(int base_pos,BitSequence& target,int base_targetPos){
  if(base_pos>=_length || base_targetPos>=target._length)return 0; //no can do, mack
  int pos=2*base_pos;//to bitspace
  int targetPos=2*base_targetPos;

  int myOffset=MODWORDSIZE(pos),tOffset=MODWORDSIZE(targetPos);
  int myFrame=pos/WORDSIZE,tFrame=targetPos/WORDSIZE;
  word myWord=readWord(myFrame),myNext=readWord(myFrame+1);
  word tWord=target.readWord(tFrame),tNext=target.readWord(tFrame+1);

  int max=min<int>(spaceRight(base_pos),target.spaceRight(base_targetPos));
  //  assert(cout << "local Spaceright: "<<spaceRight(base_pos)<< " target spaceright: "<< target.spaceRight(base_targetPos)<<endl);
  //  assert(cout << "Valid range: "<<(base_pos)<<","<<(base_pos+max)<<endl);
  if(!max)return 0; //if dead, die
  int matched=0;
  while(myFrame<word_count && tFrame<target.word_count){
    if(myOffset)
      myWord=(myWord<<myOffset) | ((myNext>>(WORDSIZE-myOffset)));
    if(tOffset)
      tWord=(tWord<<tOffset) | ((tNext>>(WORDSIZE-tOffset)));
    word res=(tWord ^ myWord);
    //    assert(cout << "Cmp right -> "<<wordToString(myWord)<<" to "<<wordToString(tWord) << " = "
    //	   << wordToString(res) << endl);
    if(res){ //sucky, not a whole match
      int highest=highestBit(res);
      matched+=WORDSIZE/2-1-(highest/2);
      break; //game over man, game over!
    }
    else
      matched+=WORDSIZE/2;
    
    myFrame++;tFrame++;
    myWord=myNext;myNext=readWord(myFrame+1);
    tWord=tNext;tNext=target.readWord(tFrame+1);
  }
  assert(matched>=0);
  if(matched>=max)
    matched=max;
  assert(matched>=0);
  return matched;
}

int BitSequence::cmpLeft(int base_pos,BitSequence& target,int base_targetPos){
  if(base_pos < 0 || base_targetPos < 0)return 0; //no can do, mack
  int pos=2*(base_pos);//to bitspace
  int targetPos=2*(base_targetPos);
  int myOffset=MODWORDSIZE(pos+2),tOffset=MODWORDSIZE(targetPos+2);
  int myFrame=pos/WORDSIZE,tFrame=targetPos/WORDSIZE;
  word myWord=readWord(myFrame),myNext=readWord(myFrame-1);
  word tWord=target.readWord(tFrame),tNext=target.readWord(tFrame-1);

  int max=min<int>(spaceLeft(base_pos),target.spaceLeft(base_targetPos));
  if(!max)return 0; //if dead, die
  int matched=0;
  while(myFrame>=0 && tFrame>=0){
    //    cout << "myOffset: "<<myOffset<<" tOffset: "<<tOffset<<endl;
    //    cout << "*Slice: "<<wordToString(myWord>>(WORDSIZE-myOffset)) << " | " << wordToString(myNext<<myOffset) << endl;
    //    cout << "& Dice: "<<wordToString(tWord) << " | "<< wordToString(tNext)<<endl;
    if(myOffset)
      myWord=(myWord>>(WORDSIZE-myOffset)) | ((myNext<<(myOffset)));
    if(tOffset)
      tWord=(tWord>>(WORDSIZE-tOffset)) | ((tNext<<(tOffset)));
    word res=(tWord ^ myWord);
    //        assert(cout << "Cmp Left <- "<<wordToString(myWord)<<" to "<<wordToString(tWord) << " = "
    //	   << wordToString(res) << endl);
    if(res){ //sucky, not a whole match
      int highest=lowestBit(res);
      matched+=(highest/2);
      break; //game over man, game over!
    }
    else
      matched+=WORDSIZE/2;
    
    myFrame--;tFrame--;
    myWord=myNext;myNext=readWord(myFrame-1);
    tWord=tNext;tNext=target.readWord(tFrame-1);
  }
  assert(matched>=0);
  if(matched>=max)
    matched=max;
  assert(matched>=0);
  return matched;
}

bool BitSequence::equal(int pos,BitSequence& target,int targetPos,BitSequence& patt){
  pos*=2;//to bitspace
  targetPos*=2; 
  int myOffset=MODWORDSIZE(pos),tOffset=MODWORDSIZE(targetPos);
  word myMask=highN(myOffset),tMask=highN(tOffset);
  int myFrame=pos/WORDSIZE,tFrame=targetPos/WORDSIZE;
  word myWord=readWord(myFrame),myNext=readWord(myFrame+1);
  word tWord=target.readWord(tFrame),tNext=readWord(tFrame+1);

  int pFrame=0;
  word pWord;

  int matched=0;
  while(myFrame<word_count && tFrame<target.word_count && pFrame<patt.word_count){
    pWord=patt.readWord(pFrame%patt.word_count); //pattern just loops for now
    if(myOffset)
      myWord=(myWord<<myOffset) | (myNext<<myOffset & ~myMask);
    if(tOffset)
      tWord=(tWord<<tOffset) | (tNext<<tOffset & ~tMask);
    word res=(tWord ^ myWord) & pWord;
    if(res){ //sucky, not a whole match
      int highest=highestBit(tWord);
      if(highest>=0) //partial match
	matched+=highest/2+1;
      break; //game over man, game over!
    }
    else
      matched+=WORDSIZE/2;
    
    myFrame++;tFrame++;pFrame++;
    myWord=myNext;myNext=readWord(myFrame+1);
    tWord=tNext;tNext=readWord(tFrame+1);
  }
  return matched>=patt._length;
}

//construct a new bitsequence from a mask
BitSequence* BitSequence::extract(int pos,BitSequence &p){
  pos*=2;//to bitspace
  int myOffset=pos%WORDSIZE;
  word myMask=lowN(myOffset);
  int myFrame=pos/WORDSIZE;
  word myWord=readWord(myFrame),myNext=readWord(myFrame+1);

  int pFrame=0;
  word pWord;
  BitSequence* idx=new BitSequence;

  while(pFrame<p.word_count){
    pWord=p.readWord(pFrame);
    if(myOffset)
      myWord=myWord<<myOffset | (myNext & myMask);

    word newWord=0;
    int newBits=0;
    for(int i=0;i<WORDSIZE;i+=2){
      if(newBits>=WORDSIZE){ //party on wayne
	idx->addWord(newWord,newBits);
	newBits=0;
      }

      if((pWord>>i) & 3){ //party on garth
	newWord=newWord<<2 | myWord & 3;
	newBits+=2;
      }
      myWord<<=2;
    }

    if(newBits)
      idx->addWord(newWord,newBits);

    //advance frames
    myFrame++;pFrame++;
    myWord=myNext;myNext=readWord(myFrame+1);
  }
  return idx;
}

void BitSequence::addWord(word w,int bits){
  _words.push_back(w);
  word_count++;
  bit_count+=bits;
  _length+=bits/2;
}

pair<int,int>* BitSequence::localRegion(int pos){ //find region local to pos
  pair<int,int> target(pos,pos); //for simplicity
  vector<pair<int,int> >::iterator
    result=lower_bound(matchRegions.begin(),matchRegions.end(),target,ltRegion());
  if(result==matchRegions.end()){
    //    assert(cout << "pos "<<pos << " not in any region"<<endl);
    return 0;
  }
  if(pos<result->first){
    //    assert(cout << "pos "<<pos<<" falls before start of next region"<<endl);
    return 0;
  }
  //  assert(cout << "region found: "<<result->first<<","<<result->second<<endl);
  return &(*result);
}

int BitSequence::spaceRight(int pos){ //find region local to pos
  pair<int,int> target(pos,pos); //for simplicity
  pair<int,int> *region(localRegion(pos));
  if(!region){
    return 0;
  }
  return region->second-pos+1;
}

int BitSequence::spaceLeft(int pos){ //find region local to pos
  pair<int,int> target(pos,pos); //for simplicity
  pair<int,int> *region(localRegion(pos));
  if(!region){
    return 0;
  }
  return pos-region->first+1;
}

//mutator
void BitSequence::invert(){ //flips to reverse comp sequence
  for(int i=0;i<word_count;i++)
    _words[i]= ~_words[i];
  //last word is tricky. all trailing elements have to be C in order
  //to not screw up dumb hashing techniques
  int last=word_count-1;
  _words[last]=_words[last] & highN(MODWORDSIZE(bit_count));
}

//init
void BitSequence::init(const string &str){
  _hashLength=0;
  _length=str.length();
  bit_count=_length*2;
  hasher=0;
  int counters[4]={0,0,0,0},counted=0;
  int bit=WORDSIZE;
  word w=0;
  bool inRegion=true;
  bool matchable=true;
  pair<int,int> region(0,0);
  for(int i=0;i<_length;i++){
    bit-=2;
    switch(str[i]){
    case 'C':case 'c': matchable=true;counters[0]++;counted++;break;
    case 'A':case 'a': w|=((word)1<<bit);counters[1]++;matchable=true;break;
    case 'T':case 't': w|=((word)2<<bit);counters[2]++;matchable=true;break;
    case 'G':case 'g': w|=((word)3<<bit);counters[3]++;matchable=true;break;
    case '1': w|=((word)3<<bit);isPattern=true;_hashLength++;matchable=true;break;
    case 'N':case 'n': matchable=false; break;
    }
    if(inRegion){ //in a matchable zone
      if(!matchable){//region end!
	region.second=i-1;
	inRegion=!inRegion;
	matchRegions.push_back(region);
      }
    }else{ //was in an N zone
      if(matchable){//region start
	region.first=i;
	inRegion=!inRegion;
      }
    }

    if(bit==0){
      _words.push_back(w);
      bit=WORDSIZE;
      w=0;
    }
  }
  if(inRegion){
    region.second=_length-1;
    matchRegions.push_back(region);
  }
  //process matchRegions into a tree for fast lookup

  if(bit<WORDSIZE)
    _words.push_back(w);
  word_count=_words.size();

  //set base frequency
  globalCounted+=counted;
  for(int b=0;b<4;b++){
    globalBaseCount[b]+=counters[b];
    //    globalBaseFreq[b]+=(double)counters[b]/(double)counted/2;
  }
}

void finishGlobalFrequencies(){
  const char *bases[4]={"C/G","A/T","T","G"};
  for(int b=0;b<2;b++){
    globalBaseFreq[b]=((double)(globalBaseCount[b]+globalBaseCount[3-b]))/(double)globalCounted/2;
    globalBaseFreq[3-b]=globalBaseFreq[b];
  }
  cout << "Global base frequencies: "<<endl;
  for(int b=0;b<2;b++){
    cout << "   "<< bases[b] << ":" << globalBaseFreq[b]/seq_count*100 << "%" << endl;
  }
}

void BitSequence::init(redi::ipstream &inf){
  string str;
  inf >> str;
  init(str);
}

void BitSequence::init(const BitSequence& a){
  _words=a._words;
  word_count=a.word_count;
  bit_count=a.bit_count;
  _length=a._length;
  _hashLength=a._hashLength;
  hasher=a.hasher;
}

//constructamalators
BitSequence::BitSequence(redi::ipstream &inf):
  isPattern(false)
{ //from input stream
  init(inf);
}

BitSequence::BitSequence(const string &str)
{ //from either GATC or 0101 (as mask: ie double each bit)
  init(str);
}

BitSequence::BitSequence(const BitSequence& a) :  //copy
  matchRegions(a.matchRegions),
  _words(a._words),
  _length(a._length),
  word_count(a.word_count),
  bit_count(a.bit_count),
  _hashLength(a._hashLength),
  isPattern(a.isPattern),
  hasher(a.hasher)
{
}

BitSequence::BitSequence() : //uninited
  _words(),_length(0),
  word_count(0),bit_count(0),
  _hashLength(-1),isPattern(false)
{}

int HashMethod::maxSources(){
  vector<int> members;
  vector<int> tmp;
  transform(inputs.begin(),inputs.end(),back_insert_iterator<vector<int> >(tmp),select1st<pair<int,int> >());
  sort(tmp.begin(),tmp.end());
  unique_copy(tmp.begin(),tmp.end(),back_insert_iterator<vector<int> >(members));
  int score=0;
  for(vector<int>::iterator i=members.begin();i!=members.end();i++){
    if(!popmap.count(*i)) //new to the area?
      score+=(popmap[*i]=popCount(seq->readWord(*i)));
    else
      score+=popmap[*i];
  }
  return score;
}

int HashMethod::entropyCheck(){
  //calculate entrop based on pattern and inputs
  int targetBits=(hash_bits+1)/2;
  entropy.clear();used.clear();
  entropy.resize(targetBits,0);
  used.resize(seq->length(),0);
  active=0;
  for(InputList::iterator i=inputs.begin();i!=inputs.end();i++){
    word w=seq->readWord(i->first);
    if(i->second > 0)
      w>>=i->second*2;
    else
      w<<=-i->second*2;
    for(unsigned b=0;b<entropy.size();b++){
      int srcbit=i->first*WORDSIZE/2+WORDSIZE/2-b-i->second-1;
      if(w & (3<<(b*2))){
	entropy[b]++;
	active|=3<<(b*2);
	if(srcbit<seq->length()) //only count stuff up till end of pat
	  used[srcbit]++;
      }
    }
  }
  
  sources=0;
  for(unsigned i=0;i<used.size();i++)
    if(used[i])
      sources++;
  assert(sources<=seq->hashLength());

  //evaluate entropy
    int sum=0,hi=entropy.front(),lo=entropy.front();
  empties=0;
  for(vector<int>::iterator i=entropy.begin();i!=entropy.end();i++){
    if(*i)
      sum+=*i;
    else
      empties++;
    hi=max<int>(hi,*i);
    lo=min<int>(lo,*i);
  }
  return sum-empties*seq->_hashLength //empties are evil
    -(hi-lo) //drawing all inputs from one source is not a good thing
    +(seq->word_count)*(targetBits-inputs.size()) //bulky outputs are bad
    +seq->_hashLength*(sources-hash_bits); //stricter enforcement of above rule
}

void BitSequence::compileHashFunc() {
  assert(isPattern);
  list<HashMethod*>  children;
  vector<int> score;
  for(int c=0;c<10;c++){ //spawn 10 kids
    children.push_back(new HashMethod(this));
  }
  
  pair<int,list<HashMethod*>::iterator > best,
    worst(children.front()->score,children.begin());
  int cycle;
  int improved=0;
  for(cycle=0;cycle<5000 || (cycle-improved)<(cycle>500 ? cycle/10:50);cycle++){ //arbitrarily large!
    //reset best/worst
    worst.first=best.first=children.front()->score;
    worst.second=best.second=children.begin();  
    //cout << "Starting iteration "<<cycle<<endl;
    for(list<HashMethod*>::iterator i=children.begin();
	i!=children.end();i++){
      if((*i)->score>best.first){
	best.first=(*i)->score;
	best.second=i;
	improved=cycle;
      }
      if((*i)->score<=worst.first){ //<= makes sure best!=worst
	worst.first=(*i)->score;
	worst.second=i;
      }
    } //resolved best and worst kid
    //    cout << "Best: "<<best.first<<" id:"<<(*best.second)->id<<" "<<**best.second<<endl;
    //    cout << "Worst: "<<worst.first<<" id:"<<(*worst.second)->id<<" "<<**best.second<<endl;

    //do we have a winner?
    if(best.first>0)
      goto CLEANUP; //oh god a goto, the joys

    assert(worst.second!=best.second);
    //kill the worst kid >_<;;
    //    cout << "Kill worst child: "<<(*worst.second)->id<<" score: "<<worst.first<<endl;
    delete *worst.second;
    children.erase(worst.second);
    //clone the best kid
    children.push_back(new HashMethod(**best.second));
    children.push_back(new HashMethod(**best.second));
    children.push_back(new HashMethod(**best.second));
    //mutate all but the best
    for(list<HashMethod*>::iterator i=children.begin();
	i!=children.end();i++)
      if(i!=best.second)
	(*i)->mutate();
    //    cout << "Done mutating."<<endl;
  }
 CLEANUP:
  cout << "Acceptable hash function found after "<<cycle-1<<" cycles."<<endl;
  hasher=*best.second;//keep the best
  for(list<HashMethod*>::iterator i=children.begin();
      i!=children.end();i++)
    if(i!=best.second)
      delete *i; //kill the rest
  hasher->finalize();
  hasher->entropyCheck();
  cout << "Best uses "<<hasher->sources<<" sources and has "<<hasher->empties<<" empties in "<<hasher->inputs.size()<<" inputs"<<endl;
  //  cout << "Pre cleanup: "<<*hasher<<endl;
  cout << "Post cleanup: "<<*hasher<<endl;
}

void HashMethod::finalize(){
  removeDuplicates();
  pruneUseless();
}

void HashMethod::removeDuplicates(){
  InputList tmp;
  sort(inputs.begin(),inputs.end());
  unique_copy(inputs.begin(),inputs.end(),insert_iterator<InputList>(tmp,tmp.begin()));
  inputs=tmp;
}

ostream& operator<<(ostream &os,const HashMethod &a){
  HashMethod::InputList::const_iterator last=a.inputs.end();
  last--;
  for(HashMethod::InputList::const_iterator i=a.inputs.begin();i!=a.inputs.end();i++){
    if(i->second > 0)
      os << "w["<<i->first<<"]>>"<<i->second;
    else if(i->second < 0)
      os << "w["<<i->first<<"]<<"<<0-i->second;
    else 
      os << "w["<<i->first<<"]";
  if(i!=last)
    os << " ^ ";
  }
  os << " {"<<a.inputs.size()<<"}";
  return os;
}

int HashMethod::align(int w){
  word pat=seq->readWord(w);
  //try all possible alignments
  multimap<int,int> score2shift;
  int bestscore=0;
  for(int shift=1-WORDSIZE/2;shift<WORDSIZE/2;shift++){
    int score;
    if(shift>0)
      score=popCount((pat>>shift*2) & ~active & hash_mask);
    else
      score=popCount((pat<<-shift*2) & ~active & hash_mask);
    bestscore=max<int>(score,bestscore);
    score2shift.insert(pair<int,int>(score,shift));
  }
  //choose randomly from among the highest scoring options
  vector<int> options;
  transform(score2shift.lower_bound(bestscore),score2shift.end(),
	    insert_iterator<vector<int> >(options,options.begin()),
	    select2nd<map<int,int>::value_type>());
  return options[random()%options.size()];
}

void HashMethod::pruneUseless(){
  vector<InputList::iterator> toErase;
  for(InputList::iterator i=inputs.begin();i!=inputs.end();i++){
    word preview=seq->readWord(i->first);
    if(i->second>=0)
      preview>>=i->second*2;
    else
      preview<<=-i->second*2;
    preview&=hash_mask;
    if(!popCount(preview))
      toErase.push_back(i);
  }
  for(vector<InputList::iterator>::iterator i=toErase.begin();
      i!=toErase.end();i++)
    inputs.erase(*i);
}

void HashMethod::mutate(){
  //  cerr << "Mutating "<<id<<" "<<*this<<" ";
  int fate=random()%100; //magical 100-sided mutation die
  if(fate<10 || maxSources()<hash_bits){//randomly add another word
    vector<int> unused;
    vector<bool> used(seq->word_count,false);
    int w; //new word is wehre?
    for(int i=0;i<(int)inputs.size();i++) //find out what's been used
      used[inputs[i].first]=true;; //incidentally all this could be kept between calls
    for(int i=0;i<seq->word_count;i++)
      if(!used[i])
	unused.push_back(i);
    if(unused.empty()) //add another word at random if boned
      w=random()%seq->word_count;
    else //if possible add a new word
      w=unused[random()%unused.size()];
    inputs.push_back(pair<int,int>(w,align(w))); //this could be done better. target a random unused bit
  }else if(fate<20 && inputs.size()>1){ //randomly kill a word
    inputs.erase(inputs.begin()+random()%inputs.size());
  }else{
    //bump a random word in a random direction
    int w=random()%inputs.size();
    int shift=inputs[w].second+random()%2 ? 1:-1;
    if(shift>=WORDSIZE/2)
      shift-=WORDSIZE; //wraparound
    else if(shift<=-WORDSIZE/2)
      shift+=WORDSIZE;
    inputs[w].second=shift;
  }
  //  removeDuplicates();
  score=entropyCheck();
  //   cerr << "->" << *this<< " new score: "<<score<<endl;
}

HashMethod::HashMethod(HashMethod &a) :
  inputs(a.inputs),seq(a.seq),score(a.score),id(nextId++),
  entropy(a.entropy),used(a.used),sources(a.sources),empties(a.empties),
  active(a.active)
{
  //  cout << a.id << " cloned => "<<id<<" ("<<*this<<")"<<endl;
}

HashMethod::HashMethod(BitSequence *s) :
  seq(s),id(nextId++),active(0)
{
  //  cout << "Brand new child "<<id;
  cout.flush();
  int w=random()%s->word_count;
  inputs.push_back(pair<int,int>(w,align(w))); //every kid starts on a random word
  score=entropyCheck();
  //  cout << " score: "<<score<<endl;
}

bool operator==(const BitSequence& a,const BitSequence& b){
  for(int i=0;i<a.word_count && i<b.word_count;i++)
    if(a.readWord(i)!=b.readWord(i))return false;
  return true;
}

Location::Location(Sequence *s,int p) :
  seqno(lookup_seqtable(s)),pos(p) { }

Location::Location(SeqIdx s,int p) :
  seqno(s),pos(p) { }

Location::Location() :
  seqno(255),pos(0) { } //this should only be used to create temp storage

Sequence* Location::seq() const{
  return seq_table[seqno];
}

BitSequence* Location::bitSeq(){return pos>0 ? seq()->fwd:seq()->rev;}

unsigned Hash::sizeAt(const HashKey key,const HashVal &a){
  //universally functionall sizeAt...
  LocList locs(seq_count);
  this->lookup(key,a,locs);
  unsigned high=0;
  for(LocList::iterator i=locs.begin();i!=locs.end();i++){
    high=max<unsigned>(i->size(),high);
  }
  return high;
}

bool ArrayHash::emptyAt(const HashKey key){
  //  return fasta[key] == NULL;
  return sizeAt(key) == 0;  // much simpler for MPI mode
}

unsigned ArrayHash::sizeAt(const HashKey key){
#ifdef MURASAKI_MPI
  if(mpi_id == 0){

    MPI_Status status;
    unsigned size, tempsize;

    // launch request & prepare for return vals
    mpi_req request;
    request.type = MURASAKI_MPI_REQ_SIZE;
    request.key  = key;

    MPI_Bcast(&request, sizeof(request), MPI_CHAR, 0, MPI_COMM_WORLD);

    // about myself
    if(fasta[key] == NULL) size = 0;
    else size = fasta[key]->size();

    // gather return vals
    for(int i=1; i<mpi_procs; i++){
      MPI_Recv(&tempsize, sizeof(unsigned), MPI_CHAR, i, 0, MPI_COMM_WORLD, &status);
      size += tempsize;
    }

    return size;
  } else
#endif
    // non-MPI version or MPI client
    if(fasta[key] == NULL) return 0;
    else return fasta[key]->size();
}

void ArrayHash::add(const HashKey key, const HashVal &a){
  if(fasta[key]==NULL) fasta[key] = new cmultiset<HashVal>;
  fasta[key]->insert(a);
}

ArrayHash::ArrayHash(BitSequence *pat) :  Hash(pat),  //holy crap C++ is cool
  activeRange(0,0)
{
  fasta=new MHash[hash_size];
  memset(fasta,0,sizeof(MHash)*hash_size);
}

void ArrayHash::clear(){
  /* // this is too slow!
  for(int i=0; i<hash_size; i++) delete fasta[i];
  memset(fasta,0,sizeof(MHash)*hash_size);
  */

  for(int i=0; i<hash_size; i++)
    if(fasta[i]!=NULL) fasta[i]->clear();
}

void ArrayHash::pickup(LocList &locList, pair<HashIte,HashIte> range){
  for(HashIte i=range.first; i!=range.second; i++)
    locList[i->seqId()].push_back(*i);
}

void ArrayHash::lookup(HashKey key, const HashVal &a, LocList &locList, bool match){

#ifdef MURASAKI_MPI
  MPI_Status status;

  // minimum manipulation overhead
  //  Location* tmpbuf;

  // minimum transfer
  SeqIdx* seqIdBuf;
  int *posBuf;
 
  unsigned tmpbufSize = 1000;

  // tmpbuf = (Location*)malloc(sizeof(Location) * tmpbufSize);
  
  seqIdBuf = (SeqIdx*)malloc(sizeof(SeqIdx) * tmpbufSize);
  posBuf   = (int*)malloc(sizeof(int) * tmpbufSize);
  if (mpi_id == 0){
    mpi_req request;
    if(match) request.type = MURASAKI_MPI_REQ_LKUP_MATCH;
    else      request.type = MURASAKI_MPI_REQ_LKUP_NOMATCH;
    request.key  = key;
    request.val  = a;

    MPI_Bcast(&request, sizeof(request), MPI_CHAR, 0, MPI_COMM_WORLD);
    
    // about myself
    if(fasta[key] != NULL){
      pair<HashIte,HashIte> range;
      if (match) range = fasta[key]->equal_range(a);
      else       range = pair<HashIte,HashIte> (fasta[key]->begin(), fasta[key]->end());
      pickup(locList, range);
    }

    // gather results
    for(int i=1; i<mpi_procs; i++){
      int s;       // message size  (in bytes)
      unsigned rs; // required size (# of Locations)
      MPI_Probe(i, 0, MPI_COMM_WORLD, &status);
      MPI_Get_count(&status, MPI_CHAR, &s);
      // rs = s/sizeof(Location);
      rs = s/sizeof(SeqIdx);

      if (rs > tmpbufSize){
	//	tmpbuf = (Location*)realloc(tmpbuf, s);
	seqIdBuf = (SeqIdx*)realloc(seqIdBuf, sizeof(SeqIdx) * rs);
	posBuf   = (int*)realloc(posBuf, sizeof(int) * rs);
  	tmpbufSize = rs;
      }
      // MPI_Recv(tmpbuf, s, MPI_CHAR, i, 0, MPI_COMM_WORLD, &status);
      MPI_Recv(seqIdBuf, rs*sizeof(SeqIdx), MPI_CHAR, i, 0, MPI_COMM_WORLD, &status);
      MPI_Recv(posBuf,   rs*sizeof(int),    MPI_CHAR, i, 0, MPI_COMM_WORLD, &status);
      
      for(unsigned j=0; j<rs; j++){
	// locList[tmpbuf[j].seqId()].push_back(tmpbuf[j]);
	Location l = Location(seqIdBuf[j], posBuf[j]);
	locList[l.seqId()].push_back(l);
      }
    }
    
    //   free(tmpbuf);
    free(seqIdBuf);
    free(posBuf);
    return;
  }
#endif

  // non-MPI version & MPI client
  if(fasta[key] != NULL){
    pair<HashIte,HashIte> range;
    if (match) range = fasta[key]->equal_range(a);
    else       range = pair<HashIte,HashIte> (fasta[key]->begin(), fasta[key]->end());
    pickup(locList, range);
  }

#ifdef MURASAKI_MPI
  // client free()
  //  free(tmpbuf);
  free(seqIdBuf);
  free(posBuf);
#endif
}

void ArrayHash::lookup(HashKey key,const HashVal &a,LocList &locList){
  lookup(key, a, locList, true);
}

void ArrayHash::lookup(HashKey key,LocList &locList){
  Location a;
  lookup(key, a, locList, false);
}

void ArrayHash::dump(ostream &os){
  HashKey key;
  unsigned s;

  for(key=0; key<(unsigned)hash_size; key++){
    s = sizeAt(key);
    //    cout << "dump: size=" << s<< endl;
    if( s==0 ) continue;
    
    os.write( (char*)&key, sizeof(HashKey)  );
    os.write( (char*)&s,   sizeof(unsigned) );

    pair <int, HashVal*> p = fasta[key]->block_read();
    HashVal* l = p.second;
    os.write( (char*)l, sizeof(HashVal) * s);
  }

  key = 0;
  s = 0;
  os.write( (char*)&key, sizeof(HashKey)  );
  os.write( (char*)&s,   sizeof(unsigned) );
}

void ArrayHash::load(istream &is){
  HashKey key;
  unsigned s;
  HashVal *buf;
  int bufsize;

  bufsize = 1024*1024;
  buf = (HashVal*)malloc(bufsize);

  do{
    is.read( (char*)&key, sizeof(HashKey) );
    is.read( (char*)&s,   sizeof(unsigned));
    if (key==0 && s==0) break;

    if(fasta[key]==NULL) fasta[key] = new cmultiset<HashVal>;
    
    int readbytes = sizeof(HashVal)*s;

    if(bufsize > readbytes){
      bufsize = readbytes;
      buf = (HashVal*)realloc(buf, bufsize);
    }
    is.read( (char*)buf, sizeof(HashVal)*s );
    fasta[key]->block_append(buf, (int)s);
  } while(true);

  free(buf);
}

void Hash::dump(ostream &os){
  HashKey key;
  unsigned s;

  for(key=0; key<(unsigned)hash_size; key++){
    unsigned c=0;
    LocList  l(seq_count); lookup(key, l);    

    s = sizeAt(key);
    if( s==0 ) continue;

    os.write( (char*)&key, sizeof(HashKey)  );
    os.write( (char*)&s,   sizeof(unsigned) );
    //    cout << "dump: size=" << s<< endl;
    for(unsigned i=0; i<l.size(); i++){
      for(list<Location>::iterator ll=l.at(i).begin(); ll!=l.at(i).end(); ll++){
	Location lll=*ll;
	os.write( (char*)&lll, sizeof(Location) );
	c++;
      }
    }
    if (s!=c){ cerr << "Hash::dump(): incorrect size?" << endl; }
  }

  key = 0;
  s = 0;
  os.write( (char*)&key, sizeof(HashKey)  );
  os.write( (char*)&s,   sizeof(unsigned) );
}

void Hash::load(istream &is){
  HashKey key;
  unsigned s;
  Location l;
  Location *buf;

  do {
    is.read( (char*)&key, sizeof(HashKey)  );
    is.read( (char*)&s,   sizeof(unsigned) );
    if( key==0 && s==0 ) break;

    buf = (Location*)malloc(sizeof(Location) * s);
    is.read( (char*)buf, sizeof(Location) * s);
    for(unsigned i=0; i<s; i++) add(key, buf[i]);
    free(buf);
  } while(true);
}

Hash::Hash(BitSequence *pat) :
  hashpat(pat)
{
}

Hash::~Hash(){}

void Hash::writeHistogram(ostream &of){
  vector<int> histo(1,0);
  for(word base=0;base<(word)hash_size;base++){ //oh this is fun...
    if(emptyAt(base))
      continue;
    usedBuckets++;
    unsigned size=sizeAt(base);
    if(size>=histo.size())
      histo.resize(size+1);
    histo[size]++;
  }
  if(of!=cerr)
    for(unsigned i=1;i<histo.size();i++){
      of << i << "\t" << histo[i]<<endl;
    }
}

void Hash::writeDetailedHistogram(ostream &of){
  for(word base=0;base<(word)hash_size;base++){ //oh this is fun...
    if(emptyAt(base))
      continue;
    unsigned size=sizeAt(base);
    if(size){
      if(opt.histogram>3){ //super freaking detailed
	LocList ll(seq_count);
	lookup(base,ll); //fetch all locations...
	vector<Location> locs;
	back_insert_iterator<vector<Location> > ii(locs);
	for(int i=0;i<seq_count;i++) 
	  copy(ll[i].begin(),ll[i].end(),ii); //throw all into big list
	sort(locs.begin(),locs.end());
	vector<Location>::iterator i=locs.begin();
	Location last(*i);
	word count=1;
	usedPats++; //we got at least 1 new one
	do{
	  i++;
	  if(i==locs.end())
	    break;
	  Location here(*i);
	  if(last<here){ //different src pattern
	    of << base << "\t" << Window(last,hashpat).prettyString() << "\t" << count << endl;
	    if(count>worstPat)
	      worstPat=count;
	    count=1;
	    last=here;
	    usedPats++;
	  }
	  count++;
	}
	while(1);
	if(count>worstPat)
	  worstPat=count;
	of << base << "\t" << Window(last,hashpat).prettyString() << "\t" << count << endl;
      }else
	of << base<<":"<<size << endl;
    }
  }
}

UsedMap::UsedMap(int count):
  used(count)
{
}

bool UsedMap::sanityCheck() {
  for(unsigned i=0;i<used.size();i++){
    assert((int)used[i].size() <= AnchorSet::_nextId);
    for(usedItree::iterator s=used[i].begin();
	s!=used[i].end();s++){
      assert(*s);
      assert((*s)->sanityCheck());
    }
  }
  return 1;
}

bool Anchor::sanityCheck() const{
  assert(parent);
  assert(bitSeqCoords(start)<seq->length());
  assert(bitSeqCoords(stop)<seq->length());
  assert(stop<0 == start<0);
  assert(start < stop);
  return 1;
}

bool AnchorSet::sanityCheck() const {
  for(unsigned i=0;i<spaces.size();i++){
    assert(spaces[i]);
    assert(spaces[i]->sanityCheck());
  }
  return 1;
}

bool ArrayHash::rawSanityCheck(){
  for(word base=0;base<(word)hash_size;base++){
    if(emptyAt(base))
      continue;
    for(HashIte i=(*fasta[base]).begin();i!=(*fasta[base]).end();i++){
      Location l(*i);
      Window w(l,hashpat);
      word here=w.hash();
      //      cout << "hash for "<<l<<" ("<<w<<"): "<<wordToString(here) << " (vs "<< wordToString(base)<<")"<<endl;
      if(here!=base){
	assert(here==base);
        return false;
      }
    }
  }
  return true;
}

bool ArrayHash::rawSanityCheck(word base){
  if((*fasta[base]).empty())
    return true;
  //  cout << "Spot checking "<<wordToString(base)<<endl;
  for(HashIte i=(*fasta[base]).begin();i!=(*fasta[base]).end();i++){
    Location l(*i);
    Window w(l,hashpat);
    word here=w.hash();
    //    cout << "  ** hash for "<<l<<" ("<<w<<"): "<<wordToString(here)<<endl;
    if(here!=base){
      assert(here==base);
      return false;
    }
  }
  return true;
}

bool Hash::sanityCheck(){
  for(word base=0;base<(word)hash_size;base++){
    if(emptyAt(base))
      continue;
    LocList loclist;
    lookup(base,loclist);
    for(unsigned s=0;s<loclist.size();s++)
      for(LocSubList::iterator i=loclist[s].begin();
	  i!=loclist[s].end();i++){
	Window w(*i,hashpat);
	word here=w.hash();
	//      cout << "hash for "<<l<<" ("<<w<<"): "<<wordToString(here) << " (vs "<< wordToString(base)<<")"<<endl;
	if(here!=base){
	  assert(here==base);
	  return false;
	}
      }
  }
  return true;
}

bool Hash::sanityCheck(word base){
  if(emptyAt(base))
    return true;
  //  cout << "Spot checking "<<wordToString(base)<<endl;
  LocList loclist;
  lookup(base,loclist);
  for(unsigned s=0;s<loclist.size();s++)
    for(LocSubList::iterator i=loclist[s].begin();
	i!=loclist[s].end();i++){
      Window w(*i,hashpat);
      word here=w.hash();
      //    cout << "  ** hash for "<<l<<" ("<<w<<"): "<<wordToString(here)<<endl;
      if(here!=base){
	assert(here==base);
      return false;
      }
    }
  return true;
}

bool UsedMap::fetchSetIds(int seq,const Space &a,set<int> &out){
  bool res=false;
  for(usedItree::range_iterator i(used[seq].in_range(a->asInterval(false,opt.joinDist)));i!=used[seq].end();i++){
    out.insert((*i)->setId());
    res=true;
  }
  return res;
}

bool UsedMap::fetchNearbySetIds(int seq,const Space &a,set<int> &out){
  bool res=false;
  for(usedItree::range_iterator i(used[seq].in_range(a->asInterval(opt.joinNeighbors,opt.joinDist)));i!=used[seq].end();i++){
    out.insert((*i)->setId());
    res=true;
  }
  return res;
}

bool UsedMap::alreadyExists(const AnchorSet& a){
  set<int> matches,temp,context;

  if(!fetchSetIds(0,a.spaces[0],context))
    return false;

  for(int i=1;i<seq_count;i++){
    if(!fetchSetIds(i,a.spaces[i],temp))
      return false;
    set_intersection(temp.begin(),temp.end(),
		     context.begin(),context.end(),
		     inserter(matches,matches.begin()));
    if(matches.empty())
      return false;
    context.clear();
    context.insert(matches.begin(),matches.end());
    matches.clear();
  }
  bool winner=false;
  //must find out if a is entirely inside of anything in context
  for(set<int>::iterator ite=context.begin();
      ite!=context.end();
      ite++){
    AnchorSet *s=AnchorSet::lookupSet(*ite);
    if(s->contains(a)){
      winner=true;
      s->hitCount++;
      if(opt.histogram>1)
	s->members[activeHash]++;
    }
  }
  return winner;
}

bool UsedMap::merge(AnchorSet& a){
  //  assert(a.sanityCheck());
  set<int> matches,temp,context;

  if(!fetchNearbySetIds(0,a.spaces[0],context))
    return false;

  for(int i=1;i<seq_count;i++){
    if(!fetchNearbySetIds(i,a.spaces[i],temp))
      return false;
    set_intersection(temp.begin(),temp.end(),
		     context.begin(),context.end(),
		     inserter(matches,matches.begin()));
    if(matches.empty())
      return false;
    context.clear();
    context.insert(matches.begin(),matches.end());
    matches.clear();
  }
  //context now contains all anchorsets which exist in all seqs
  //(but might not overlap!!!)
  assert(a.sanityCheck());
  for(set<int>::iterator i=context.begin();
      i!=context.end();
      i++){
    AnchorSet *neighbor=AnchorSet::lookupSet(*i);
    assert(neighbor->sanityCheck());
    if(!a.overlaps(*neighbor))
      continue;
    //    if(verbose)cout << "Merging: "<<a<<" with "<< *neighbor<<endl;
    //    assert(sanityCheck());
    //    assert(a.sanityCheck());
    a.coalesce(*neighbor);
    //    assert(sanityCheck());
    //    assert(a.sanityCheck());
    remove(neighbor);
    //    if(verbose)cout << "Result: "<<a<<endl;
    assert(sanityCheck());
    assert(a.sanityCheck());
  }
  add(a);
  return true;
}

void UsedMap::remove(AnchorSet *a){
  //  assert(sanityCheck());
  for(unsigned i=0;i<a->spaces.size();i++){
    assert(cerr << "Erasing seq "<<i<<" - "<<*(a->spaces[i])<<" from: "<<used[i]<<endl);
    used[i].erase(a->spaces[i]->mapite);
    //    assert(sanityCheck());
  }
  //  assert(sanityCheck());
  delete a;
  //  assert(sanityCheck());
}

void UsedMap::add(AnchorSet& a){
  //  a.sanityCheck();
  //  if(alreadyExists(a)) //this check is already done on murasaki side
  //    return;
  
  for(unsigned i=0;i<a.spaces.size();i++){
    a.spaces[i]->mapite=used[i].insert(a.spaces[i]->asInterval(),a.spaces[i]); //for later deletion
    assert(cerr << "Added "<<*a.spaces[i]<<" to tree"<<i<<": "<<used[i]<<endl);
  }
    
}

bool AnchorSet::contains(const AnchorSet &a){
  for(unsigned i=0;i<spaces.size();i++){
    if(!spaces[i]->contains(*a.spaces[i]))
      return false;
  }
  return true;
}

bool AnchorSet::overlaps(const AnchorSet &a){
  for(unsigned i=0;i<spaces.size();i++){
    if(spaces[i]->cmp(*a.spaces[i]))
      return false;
  }
  return true;
}

bool Anchor::contains(const Anchor &a){
  return start<=a.start && stop>=a.stop;
}

void AnchorSet::coalesce(const AnchorSet &a){
  for(unsigned i=0;i<spaces.size();i++)
    spaces[i]->coalesce(*a.spaces[i]);
  //mmmm tasty (metadata - omph omph!)
  hitCount+=a.hitCount; 
  if(opt.histogram>1)
    for(HashCount::const_iterator i=a.members.begin();i!=a.members.end();i++)
      members[i->first]+=i->second;
}

void Anchor::coalesce(const Anchor &a){
  start=start<=a.start ? start:a.start;
  stop=stop>=a.stop ? stop:a.stop;
}


AnchorSet::AnchorSet():
  setId(_nextId++),hitCount(1)
{
  setMap[setId]=this;
  if(opt.histogram>1)
    members[activeHash]++;
}

void AnchorSet::add(Space a){
  //  assert(a);
  spaces.push_back(a);
  mclass|=a->seq->mclass();
}

void AnchorSet::add(HashVal &a,int length){
  length--;
  spaces.push_back(new Anchor(a.pos>0 ? a.pos:a.pos-length,a.pos>0 ? a.pos+length:a.pos,this,a.seq()));
  //  assert(spaces.back());
  mclass|=a.seq()->mclass();
}

AnchorSet::~AnchorSet(){
  for(unsigned i=0;i<spaces.size();i++){
    assert(spaces[i]);
    delete spaces[i];
  }
  setMap.erase(setId);
}

Anchor::Anchor(int _start,int _stop,AnchorSet *_parent,Sequence *_seq):
  start(_start),stop(_stop),
  parent(_parent),seq(_seq),
  mapite(0)
{
  assert(parent);
}

int Anchor::setId() const {
  assert(parent);
  return parent->setId;
}

int Anchor::length() const {
  return stop-start+1;
}

int Anchor::cmp(const Anchor &a) const {
  //  assert(sanityCheck());
  //  assert(a.sanityCheck());
  if(start<=a.stop)
    if(a.start<=stop)
      return 0; //overlaps at least partially
    else
      return 1; //obj comes completely after this
  else
    return -1; //obj comes completely before this
}

bool operator<(const Anchor &a,const Anchor &b){
  return a.cmp(b)>0;
}

bool operator<(const Location &a,const Location &b){
  Location copya(a),copyb(b);
  Window wina(copya,mfh->hashpat),winb(copyb,mfh->hashpat);
  //  assert(cout << "Comparing "<<wina.prettyString()<<" to "<<winb.prettyString()
  //	 << " : "<< (wina<winb ? "<":">=")<<endl);
  //mm construct two temporary windows? delicious
  return wina<winb;
}

bool operator<(const Window &a,const Window &b){
  for(int frame=0;frame<a.frames;frame++){
    word pword=a.pat->readWord(frame);
    word resa=a.buffer[frame] & pword;
    word resb=b.buffer[frame] & pword;
    if(resa<resb)
      return 1;
    if(resb<resa)
      return 0;
  }
  return 0;
}

void AnchorSet::extend(){
  BitSequence *mySeq=spaces[0]->stop>0 ? spaces[0]->seq->fwd :
    spaces[0]->seq->rev,*tSeq;
  int mstart=bitSeqCoords(spaces[0]->start),tstart;
  int mstop=bitSeqCoords(spaces[0]->stop),tstop;
  int right=mySeq->length()-mstop,
    left=mstart;
  assert(mstart>=0);
  for(unsigned i=1;i<spaces.size();i++){
    tSeq=spaces[i]->stop>0 ? spaces[i]->seq->fwd :
      spaces[i]->seq->rev;
    tstart=bitSeqCoords(spaces[i]->start);
    tstop=bitSeqCoords(spaces[i]->stop);
    if(spaces[i]->stop<0)
      swap(tstart,tstop); //sneaky devils

    assert(tstart>=0);
    
    right=min<int>(tSeq->length()-tstop-1,min<int>(right,mySeq->cmpRight(mstop+1,*tSeq,tstop+1)));
    left=min<int>(tstart,min<int>(left,mySeq->cmpLeft(mstart-1,*tSeq,tstart-1)));
  }

  if(right || left){
    assert(cout << *this << "-> Extending "<<mstart<<","<<mstop<<" by "<<left<<","<<right<<" to..."<<endl);
    assert(sanityCheck());
    if(right)
      for(unsigned i=0;i<spaces.size();i++)
	if(spaces[i]->stop>0)
	  spaces[i]->stop+=right;
	else
	  spaces[i]->start-=right;
    if(left)
      for(unsigned i=0;i<spaces.size();i++)
	if(spaces[i]->start>0)
	  spaces[i]->start-=left;
	else
	  spaces[i]->stop+=left;
    assert(cout << *this<<"<------"<<endl);
  }else{
    assert(cout << *this << "\\\\// Not extending"<<endl);
  }
  assert(sanityCheck());
}

bool operator==(const Location &a,const Location &b) {
  BitSequence *mySeq=a.pos>0 ? a.seq()->fwd : a.seq()->rev,
    *tSeq=b.pos>0 ? b.seq()->fwd : b.seq()->rev;
  int mstop=bitSeqCoords(a.pos),
    tstop=bitSeqCoords(b.pos);
  return mySeq->equal(mstop,*tSeq,tstop,*mfh->hashpat);
}

ostream& operator<<(ostream &os, const Window &a){
  for(int i=0;i<a.frames-1;i++){
    os << wordToString(a.buffer[i]) << " ";
  }
  os << "+" << wordToString(a.buffer[a.frames-1]);
  return os;
}

ostream& operator<<(ostream &os,const BitSequence &a){
  if(a.isPattern)
    return os<<a.asPattern();
  else
    return os << a.asString();
}

ostream& operator<<(ostream &os,const Location &a){
  return os << "Location: ("<< a.seq()->getId()<<","<<a.seq()->extCoords(a.pos)<<")";
}

void BitSequence::maskString(string &str) const{
  int j=0;
  for(int i=0;i<word_count;i++){
    word w=readWord(i);
    word mask=highN(2);
    while(mask){
      if(!(w & mask))
	str[j]='.';
      j++;
      mask>>=2;
    }
  }
}

string Window::toRawString(){
  string ret;
  for(unsigned i=0;i<buffer.size();i++)
    ret+=wordToString(buffer[i]);
  ret.resize(pat->length());
  return ret;
}

string Window::prettyString(){
  string ret(toRawString());
  pat->maskString(ret);
  size_t len=ret.length(),patlen=min(pat->length(),hash_bits/2);
  switch(opt.quick_hash){
  case 2:
    for(size_t i=patlen;i<len;i++)
      ret[i]=tolower(ret[i]);
    break;
  case 1:
    for(size_t i=WORDSIZE/2;i<len;i++)
      ret[i]=tolower(ret[i]);
    break;
  case 3:
    for(size_t i=0;i<len;i++)
      if(!pat->hasher->used[i])
	ret[i]=tolower(ret[i]);
    break;
  }
  return ret;
}

void Window::slide(){
  pos+=2;
  int i;
  for(i=0;i<frames-1;i++){
    buffer[i]=(buffer[i]<<2) | (buffer[i+1]>>(WORDSIZE-2));
  }
  buffer[i]<<=2;

  //this should be done at the end. thus frame is always correct and no buffers are ever at empty
  if(pos/WORDSIZE!=frame){
    frame=pos/WORDSIZE;
    buffer[frames-1]=seq->readWord(frame+frames-1);
  }
}

void Window::slideWord(){
  //update to new positions
  pos+=WORDSIZE;
  frame++;

  int i;
  for(i=0;i<frames-1;i++){
    buffer[i]=buffer[i+1];
  }
  //[i] is now the last buffer
  buffer[i]=seq->readWord(frame+frames-1);
  int offset=MODWORDSIZE(pos);
  if(offset) //[i-1] needs to get the offset bits that had been chopped off of the previous [i]
    buffer[i-1]|=buffer[i]>>(WORDSIZE-offset);
  //and [i] needs to not duplicate those on a subsequent slide...
  buffer[i]<<=offset;
}

void Window::slide(int dist){ //bah. this function isnt actually used so, dont care that it's not as efficient as possible
  while(dist>WORDSIZE){
    slideWord();
    dist-=WORDSIZE;
  }

  while(dist>0){
    slide();
    dist-=2;
  }
}

void Window::initslide(int dist){
  assert(dist<=WORDSIZE-(MODWORDSIZE(pos)));
  pos+=dist;
  int i;
  for(i=0;i<frames-1;i++){
    buffer[i]=(buffer[i]<<dist) | (buffer[i+1]>>(WORDSIZE-dist));
  }
  buffer[i]<<=dist;

  if(pos/WORDSIZE!=frame){
    frame=pos/WORDSIZE;
    buffer[frames-1]=seq->readWord(frame+frames-1);
  }
}

bool Window::equals(const Window &a) const {
  assert(pat==a.pat); //dont be dumb
  if(pos+pat->bit_count>seq->bit_count) //overruns end of seq
    return false;
  for(int i=0;i<frames-1;i++){
    word result=(buffer[i] & pat->readWord(i)) ^ (a.buffer[i] & pat->readWord(i));
    if(result)
      return false; //a bit's amiss!
  }
  return true;
}

inline word Window::eqbases(const Window &a) const {
  return ~(buffer[0] ^ a.buffer[0]);
}

Window::Window(BitSequence *_seq,int _pos,BitSequence *_pat):
  seq(_seq),pat(_pat),pos(_pos*2),
  frames(_pat->word_count+1),buffer(frames)
{
  frame=pos/WORDSIZE;
  pos=frame*WORDSIZE;
  for(int i=0;i<frames;i++){
    buffer[i]=seq->readWord(frame+i);
  }
  if(pos!=_pos*2)
    initslide(MODWORDSIZE(_pos*2));//scuttle right the leftover bits
}

Window::Window(Location &v,BitSequence *p):
  seq(v.bitSeq()),pat(p),pos(bitSeqCoords(v.pos)*2),
  frames(p->word_count+1),buffer(frames)
{
  int stop=pos;
  frame=pos/WORDSIZE;
  pos=frame*WORDSIZE;
  for(int i=0;i<frames;i++)
    buffer[i]=seq->readWord(frame+i);
  if(pos!=stop)
    initslide(MODWORDSIZE(stop)); //scuttle right the leftover bits
}

Window::Window(Anchor *a,BitSequence *p):
  seq(a->start>0 ? a->seq->fwd:a->seq->rev),pat(p),pos(bitSeqCoords(a->start>0 ? a->start:a->stop)*2),
  frames(p->word_count+1),buffer(frames)
{
  int stop=pos;
  frame=pos/WORDSIZE;
  pos=frame*WORDSIZE;
  for(int i=0;i<frames;i++)
    buffer[i]=seq->readWord(frame+i);
  if(pos!=stop)
    initslide(MODWORDSIZE(stop)); //scuttle right the leftover bits
}
  
word Window::hash(){
  word hash=0;
  int filled=0,bit=WORDSIZE-2,frame=0;
  int patpos=0;
  switch(opt.quick_hash){
  case 2: //naive hash: use first hash_bits bits
  while(filled<WORDSIZE && patpos<pat->bit_count){
    if(bit<0){
      frame++;
      bit=WORDSIZE-2;
    }
    if(pat->readCode(patpos)){
      hash=hash<<2 | ((buffer[frame]>>bit) & 3);
      filled+=2;
    }
    patpos+=2;
    bit-=2;
  }
  break;
  case 1: //use first word only (usually very bad waste of hash)
    hash=buffer[0] & pat->readWord(0);
    break;
  case 3: //adaptive hasher. frekain' smart (usually...)
    word patword=0;
    int lastpati=-1;
    for(HashMethod::InputList::iterator i=pat->hasher->inputs.begin();
	i!=pat->hasher->inputs.end();i++){
      if(i->first!=lastpati)
	patword=pat->readWord(i->first);
      if(i->second>=0)
	hash^=(buffer[i->first] & patword)>>i->second*2;
      else
	hash^=(buffer[i->first] & patword)<<-i->second*2;
      lastpati=i->first;
    }
    break;
  }

  return hash & hash_mask;
}

ostream& operator<<(ostream &os, const AnchorSet &a){
  return os << a.asString();
}

string UsedMap::asString() {
  string ret;
  for(usedItree::iterator i=used[0].begin();
      i!=used[0].end();i++){
    assert((*i)->parent);
    ret+= (*i)->parent->asString() + "\n";
  }
  return ret;
}

ostream& operator<<(ostream &os,const pair<word,int> &p){
  return os << p.first << ":" << p.second;
}

ostream& UsedMap::saveDetails(ostream &os,ostream &bitos){
  for(usedItree::iterator i=used[0].begin();
      i!=used[0].end();i++){
    assert((*i)->parent);
    AnchorSet *&s=(*i)->parent;
    if(s->hitCount>=opt.hitfilter){
      os << s->asString() << "\t" << s->hitCount;

      if(opt.histogram>1){
	os << "\t" << s->uniqueness() << "\t";
	AnchorSet::HashCount::iterator last=s->members.end();
	assert(!s->members.empty());
	last--;
	for(AnchorSet::HashCount::iterator i=s->members.begin();i!=last;i++)
	  os << i->first << ":" << i->second <<",";
	os << last->first << ":" << last->second;
      }

      if(opt.bitscore && bitos!=cerr){
	pair<int,int> scores(s->bitscore());
	bitos << scores.first << "\t" << scores.second << endl;
      }
      os << endl;
    }
  }
  return os;
}

int UsedMap::count() const {
  return used[0].size();
}

void UsedMap::writeOut(ostream &os){
  for(usedItree::iterator i=used[0].begin();
      i!=used[0].end();i++){
    assert((*i)->parent);
    if((*i)->parent->hitCount>=opt.hitfilter)
      os << (*i)->parent->asString() << "\n";
  }
  //as these output sets can get huge, this might be a bad idea
  //  return os << asString();
}

ostream& operator<<(ostream &os, const Anchor &a){
  return os << a.asString();
}

string AnchorSet::asString() const {
  string ret;
  for(unsigned i=0;i<spaces.size();i++){
    assert(spaces[i]);
    ret+=spaces[i]->asString();
    if(i!=spaces.size()-1)
      ret+="\t";
  }
  return ret;
}

string Anchor::asString() const {
  if(start<0) //gotta flip
    return dstring(seq->extCoords(stop))+string("\t")+
    dstring(seq->extCoords(start))+string("\t")+
    string(((start<0) != (stop<0)) ? "?":(start>=0 ? "+":"-"));
  else return dstring(seq->extCoords(start))+string("\t")+
    dstring(seq->extCoords(stop))+string("\t")+
    string(((start<0) != (stop<0)) ? "?":(start>=0 ? "+":"-"));
}

AnchorSet* AnchorSet::lookupSet(int id){
  //  assert(setMap.count(id));
  return setMap[id];
}

double AnchorSet::uniqueness() const {
  double sum=0;
  for(HashCount::const_iterator i=members.begin();i!=members.end();i++){
    // wi = tfi*log(D/dfi) by the money
    if(!mfh->sizeAt(i->first)){
      cout << "What the monkey? "<<(i->first)<<" isn't in hash?\n";
    }
    sum+=i->second*log((double)usedBuckets/(double)mfh->sizeAt(i->first));
  }
  return sum;
}

pair<double,int> AnchorSet::entropy() const {
  double ret=0;
  vector<Window> windows;
  int shortest=0,shortlen=spaces[0]->length();
  for(int i=0;i<(int)spaces.size();i++){
    windows.push_back(Window(spaces[i],mfh->hashpat));
    if(shortlen>spaces[i]->length()){
      shortest=i;
      shortlen=spaces[i]->length();
    }
  }

  int counters[4],offset;
  for(int pos=0;pos<shortlen;){
    for(offset=WORDSIZE-2;offset>=0 && pos<shortlen;pos++,offset-=2){
      memset(counters,0,sizeof(int)*4);
      
      for(uint s=0;s<spaces.size();s++)
	counters[(windows[s].firstWord()>>offset) & 3]++;
      
      //calculate (and add in) entropy for this column
      for(int b=0;b<4;b++)
	if(counters[b]){
	  double count=(double)counters[b];
	  ret+=count*log(count/globalBaseFreq[b]);
	}
    }
    for(vector<Window>::iterator ite=windows.begin();ite!=windows.end();ite++)
      ite->slideWord();
  }
  return pair<double,int>(ret,shortlen);
}

pair<int,int> AnchorSet::bitscore() const {
  vector<Window> windows;
  int shortest=0,shortlen=spaces[0]->length();
  for(int i=0;i<(int)spaces.size();i++){
    windows.push_back(Window(spaces[i],mfh->hashpat));
    if(shortlen>spaces[i]->length()){
      shortest=i;
      shortlen=spaces[i]->length();
    }
  }
  int eqbits=0;
  for(int pos=0;pos<shortlen;pos+=WORDSIZE/2){
    word ret=~0;
    //compare everything to ref seq. store in ret
    for(int comp=0;comp<(int)windows.size();comp++){
      if(comp==shortest)continue;
      word eq=windows[comp].eqbases(windows[shortest]);
      if(pos+WORDSIZE/2>spaces[comp]->length())
	eq=(eq & (highN(2*(spaces[comp]->length()-pos))));
      eq=(((eq & alternating_mask)<<1) & eq);
      ret&=eq;
    }
    //count ret and add it up!
    eqbits+=popCount(ret);
    
    //shuffle everybody
    for(vector<Window>::iterator ite=windows.begin();ite!=windows.end();ite++)
      ite->slideWord();
  }
  return pair<int,int>(eqbits,shortlen);
}

string hashCacheName(string name, BitSequence *pat){
  ostringstream os;
  
  char *n = strdup(name.c_str());
  string fr;
  if ( !opt.skipFwd && !opt.skipRev ) fr = "both";
  if ( !opt.skipFwd &&  opt.skipRev ) fr = "fwd";
  if (  opt.skipFwd && !opt.skipRev ) fr = "rev";

  string rm;
  if ( opt.repeatMask ) rm="r"; else rm="nr";
					
    
  for(unsigned int i=0; i<strlen(n); i++)
    if(n[i]=='/' || n[i]=='.') n[i]='-';
    
  os << opt.cache_dir << "mc-" << n << "-";
  for(unsigned int p=0; p<pat->words().size(); p++)
    os << hex << pat->words().at(p);
  os << "-" << fr;
  os << "-" << rm;
  os << "-" << dec << opt.rand_seed;
  os << "-" << hash_bits;
  free(n);
    
  return os.str();

}

void mpi_client_mode(){
#ifdef MURASAKI_MPI

  cout << "MPI client #" << mpi_id << " is now ready!" << endl;

  mpi_req request;

  // minimum data manipulation overhead
  // Location* locBuf;

  // minimum data transfer
  SeqIdx* seqIdBuf;
  int* posBuf;

  unsigned locBufSize = 1000;
  unsigned s;

  // locBuf = (Location*)malloc(sizeof(Location) * locBufSize);
  
  seqIdBuf = (SeqIdx*)malloc(sizeof(SeqIdx) * locBufSize);
  posBuf   = (int*)malloc(sizeof(int) * locBufSize);
    
  do {
    MPI_Bcast(&request, sizeof(request), MPI_CHAR, 0, MPI_COMM_WORLD);
    if(request.type == MURASAKI_MPI_REQ_QUIT) break;
    
    switch( request.type ){
    case MURASAKI_MPI_REQ_SIZE:
      s = mfh->sizeAt(request.key);
      MPI_Send(&s, sizeof(s), MPI_CHAR, 0, 0, MPI_COMM_WORLD);
      break;

    case MURASAKI_MPI_REQ_LKUP_NOMATCH:
    case MURASAKI_MPI_REQ_LKUP_MATCH:
      LocList locList(seq_count);

      if( request.type == MURASAKI_MPI_REQ_LKUP_NOMATCH )
	mfh->lookup(request.key, locList);
      else 
	mfh->lookup(request.key, request.val, locList);

      // flatten locList
      unsigned crs=0; // current required size
      unsigned l=0;
      for(unsigned j=0; j<locList.size(); j++){
	crs += locList[j].size();
	if ( crs > locBufSize ){ 
	  // locBuf = (Location*)realloc(locBuf, sizeof(Location) * crs);
	  seqIdBuf = (SeqIdx*)realloc(seqIdBuf, sizeof(SeqIdx) * locBufSize);
	  posBuf   = (int*)realloc(posBuf, sizeof(int) * locBufSize);
	  locBufSize = crs; }

	for(list<Location>::iterator k=locList[j].begin(); k!=locList[j].end(); k++){
	  // locBuf[l] = *k;
	  seqIdBuf[l] = k->seqno;
	  posBuf[l] = k->pos;
	  l++;
	}
      }

      // MPI_Send((void *)locBuf, sizeof(Location)*crs, MPI_CHAR, 0, 0, MPI_COMM_WORLD);
      MPI_Send((void*)seqIdBuf, sizeof(SeqIdx)*crs, MPI_CHAR, 0, 0, MPI_COMM_WORLD);
      MPI_Send((void*)posBuf,   sizeof(int)*crs,    MPI_CHAR, 0, 0, MPI_COMM_WORLD);

      break;
    }

  } while (true);

  cout << "MPI client #" << mpi_id << " terminating." << endl;
#endif
}
